'' =================================================================================================
''
''   File....... jm_background_timer.spin
''   Purpose....
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2019 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started....
''   Updated.... 29 DEC 2019
''
'' =================================================================================================


con { timing }

  _clkmode = xtal1 + pll16x                                     
  _xinfreq = 5_000_000                                          ' use 5MHz crystal

  CLK_FREQ = (_clkmode >> 6) * _xinfreq                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us

  BR_TERM  = 115_200                                            ' terminal baud rate
  

con { io pins }                                                  
                                                                 
  RX1   = 31  { I }                                             ' serial / programming
  TX1   = 30  { O }                                            
                                                             
  SDA1  = 29  { I/O }                                           ' i2c / eeprom
  SCL1  = 28  { I/O }

  LED27 = 27 { O }                                              ' led on PAB and FLiP
  

con

  #true,  ON, OFF 
  #false, NO, YES


obj

' main                                                          ' * master Spin cog
  time       : "jm_time_80"                                     '   timing and delays (80MHz)
  prng       : "jm_prng"                                        '   pseudo-random number generator
  io         : "jm_io_basic"                                    '   essential io
  term       : "jm_fullduplexserial"                            ' * serial IO for terminal
' background                                                    ' * background timing & processes                                                    
                                                                 
' * uses cog when loaded

                                                                 
dat { pre-initialized }


var { globals }                                                              
                                                                 
                                                                 
pub main | last                                                        
                                                                 
  setup

  last := -1

  term.tx(term#CLS)                                             ' clear terminal to start                                            
                                                             
  repeat
    time.pause(5)
    term.tx(term#HOME)
    term.dec(millis)
    term.tx(13)
                                                           
    if (rtc.byte[RTC_S] <> last)                                ' change?
      show_clock                                                ' show rtc
      term.tx(13)
      term.dec(millis / 1000)                                   ' show background timer
      term.tx(13)
      term.dec(time.seconds)                                    ' show elapsed seconds from time object
      last := rtc.byte[RTC_S]


pub show_clock | clock

  clock := rtc                                                  ' make copy

  dec2(clock.byte[RTC_H])
  term.tx(":")
  dec2(clock.byte[RTC_M])
  term.tx(":")
  dec2(clock.byte[RTC_S])


pub dec2(value)

  if ((value < 0) or (value > 99))
    term.txn("?", 2)
  else
    term.tx("0" + value  / 10)
    term.tx("0" + value // 10)                                                   
                                                                 
                                                                 
pub setup                                                        
                                                                 
'' Setup IO and objects for application                          
                                                                 
  time.start                                                    ' setup timing & delays
                                                                 
  prng.seed(-cnt, -cnt ~> 2, $CAFE_BABE, cnt <- 2, cnt)         ' seed randomizer                                                                           

  io.start(0, 0)                                                ' clear all pins (master cog)
                                                                 
  term.start(RX1, TX1, %0000, BR_TERM)                          ' start serial for terminal *

  bcog := cognew(background, @bstack) + 1                       ' start background cog *

                                                              
con { 'background' processes }

var

  long  bcog                                                    ' cog running background
  long  bstack[32]                                              ' stack for background Spin cog

  long  millis                                                  ' global timer


pri background | t

  bkg_setup                                                     ' setup hardware used by "background"

  t := cnt                                                      ' sync 1ms timer  
  repeat
    waitcnt(t += MS_001)                                        ' set loop time to 1ms
    ++millis                                                    ' update millis timer
    heartbeat
    update_rtc

    ' run other "background" processes
    ' -- total of all must be under 1ms


pri bkg_setup

  io.low(LED27)


pri heartbeat

  if ((millis // 1000) < 500)                                   ' led on first half of each second
    outa[LED27] := 1
  else
    outa[LED27] := 0


con

  #0, RTC_S, RTC_M, RTC_H, RTC_D                                ' byte offsets into rtc
  

var

  long  rtcmillis                                               ' millis time for rtc regs
  long  rtc                                                     ' rtc regs
  long  rtcreset                                                ' rtc reset flag


pri update_rtc

  if (rtcreset)                                                 ' reset?
    longfill(@rtcmillis, 0, 3)                                  '  reset millis, regs, and flag

  else
    if (++rtcmillis == 1000)                                    ' has 1 second elapsed?
      rtcmillis := 0
      if (++rtc.byte[RTC_S] == 60)
        rtc.byte[RTC_S] := 0
        if (++rtc.byte[RTC_M] == 60)
          rtc.byte[RTC_M] := 0
          if (++rtc.byte[RTC_H] == 24)
            rtc.byte[RTC_H] := 0
            if (++rtc.byte[RTC_D] == 7)
              rtc.byte[RTC_D] := 0

                          
con { license }                                                  

{{

  MIT License

  Copyright (c) 2019 Jon McPhalen

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:
  
  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.
  
  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

}}     